@@ -2,7 +2,10 @@ import click
 import neuro_sdk
 import signal
 import sys
-from typing import List, Optional, Tuple
+from datetime import datetime, timezone
+from dateutil.parser import isoparse
+from neuro_cli.parse_utils import parse_timedelta
+from typing import List, Optional, Sequence, Tuple
 
 from neuro_flow.batch_executor import ExecutorData
 from neuro_flow.batch_runner import BatchRunner
@@ -12,10 +15,11 @@ from neuro_flow.cli.click_types import (
     BATCH_OR_ALL,
     FINISHED_TASK_AFTER_BAKE,
 )
-from neuro_flow.cli.utils import argument, option, wrap_async
-from neuro_flow.storage import FSStorage, NeuroStorageFS, Storage
+from neuro_flow.cli.utils import argument, option, resolve_bake, wrap_async
+from neuro_flow.storage import APIStorage, NeuroStorageFS, Storage
 from neuro_flow.types import LocalPath
 
+from ..parser import parse_bake_meta
 from .root import Root
 
 
@@ -30,13 +34,37 @@ else:
 @click.option(
     "--param", type=(str, str), multiple=True, help="Set params of the batch config"
 )
+@click.option(
+    "-n",
+    "--name",
+    metavar="NAME",
+    type=str,
+    help="Optional bake name",
+    default=None,
+)
+@click.option(
+    "--meta-from-file",
+    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
+    help="File with params for batch.",
+)
+@click.option(
+    "-t",
+    "--tag",
+    metavar="TAG",
+    type=str,
+    help="Optional bake tag, multiple values allowed",
+    multiple=True,
+)
 @argument("batch", type=BATCH)
 @wrap_async()
 async def bake(
     root: Root,
     batch: str,
     local_executor: bool,
+    meta_from_file: Optional[str],
     param: List[Tuple[str, str]],
+    name: Optional[str],
+    tag: Sequence[str],
 ) -> None:
     """Start a batch.
 
@@ -45,12 +73,22 @@ async def bake(
     async with AsyncExitStack() as stack:
         client = await stack.enter_async_context(neuro_sdk.get())
         storage: Storage = await stack.enter_async_context(
-            FSStorage(NeuroStorageFS(client))
+            APIStorage(client, NeuroStorageFS(client))
         )
         runner = await stack.enter_async_context(
-            BatchRunner(root.config_dir, root.console, client, storage)
+            BatchRunner(root.config_dir, root.console, client, storage, root)
+        )
+        params = {key: value for key, value in param}
+        if meta_from_file is not None:
+            bake_meta = parse_bake_meta(LocalPath(meta_from_file))
+            params = {**bake_meta, **params}
+        await runner.bake(
+            batch_name=batch,
+            local_executor=local_executor,
+            params=params,
+            name=name,
+            tags=tag,
         )
-        await runner.bake(batch, local_executor, {key: value for key, value in param})
 
 
 @click.command(hidden=True)
@@ -82,33 +120,70 @@ async def execute(
         data = ExecutorData.parse(executor_data)
         client = await stack.enter_async_context(neuro_sdk.get())
         storage: Storage = await stack.enter_async_context(
-            FSStorage(NeuroStorageFS(client))
+            APIStorage(client, NeuroStorageFS(client))
         )
         runner = await stack.enter_async_context(
-            BatchRunner(root.config_dir, root.console, client, storage)
+            BatchRunner(root.config_dir, root.console, client, storage, root)
         )
         await runner.process(data)
 
 
 @click.command()
+@click.option(
+    "-t",
+    "--tag",
+    metavar="TAG",
+    type=str,
+    help="Filter out bakes by tag (multiple option)",
+    multiple=True,
+)
+@option(
+    "--since",
+    metavar="DATE_OR_TIMEDELTA",
+    help="Show bakes created after a specific date (including). "
+    "Use value of format '1d2h3m4s' to specify moment in "
+    "past relatively to current time.",
+)
+@option(
+    "--until",
+    metavar="DATE_OR_TIMEDELTA",
+    help="Show bakes created before a specific date (including). "
+    "Use value of format '1d2h3m4s' to specify moment in "
+    "past relatively to current time.",
+)
+@option(
+    "--recent-first/--recent-last",
+    is_flag=True,
+    default=False,
+    help="Show newer bakes first or last",
+)
 @wrap_async()
 async def bakes(
     root: Root,
+    tag: Sequence[str],
+    since: Optional[str],
+    until: Optional[str],
+    recent_first: bool,
 ) -> None:
     """List existing bakes."""
     async with AsyncExitStack() as stack:
         client = await stack.enter_async_context(neuro_sdk.get())
         storage: Storage = await stack.enter_async_context(
-            FSStorage(NeuroStorageFS(client))
+            APIStorage(client, NeuroStorageFS(client))
         )
         runner = await stack.enter_async_context(
-            BatchRunner(root.config_dir, root.console, client, storage)
+            BatchRunner(root.config_dir, root.console, client, storage, root)
+        )
+        await runner.list_bakes(
+            tags=set(tag),
+            since=_parse_date(since),
+            until=_parse_date(until),
+            recent_first=recent_first,
         )
-        await runner.list_bakes()
 
 
 @click.command()
-@argument("bake_id", type=BAKE)
+@argument("bake", type=BAKE)
 @option(
     "-a",
     "--attempt",
@@ -143,7 +218,7 @@ async def bakes(
 @wrap_async()
 async def inspect(
     root: Root,
-    bake_id: str,
+    bake: str,
     attempt: int,
     output_graph: Optional[str],
     dot: bool,
@@ -157,15 +232,17 @@ async def inspect(
     async with AsyncExitStack() as stack:
         client = await stack.enter_async_context(neuro_sdk.get())
         storage: Storage = await stack.enter_async_context(
-            FSStorage(NeuroStorageFS(client))
+            APIStorage(client, NeuroStorageFS(client))
         )
         runner = await stack.enter_async_context(
-            BatchRunner(root.config_dir, root.console, client, storage)
+            BatchRunner(root.config_dir, root.console, client, storage, root)
         )
         if output_graph is not None:
             real_output: Optional[LocalPath] = LocalPath(output_graph)
         else:
             real_output = None
+        bake_id = await resolve_bake(bake, project=runner.project_id, storage=storage)
+
         await runner.inspect(
             bake_id,
             attempt_no=attempt,
@@ -177,7 +254,7 @@ async def inspect(
 
 
 @click.command()
-@argument("bake_id", type=BAKE)
+@argument("bake", type=BAKE)
 @argument("task_id", type=FINISHED_TASK_AFTER_BAKE)
 @click.option(
     "-a",
@@ -198,7 +275,7 @@ async def inspect(
 @wrap_async()
 async def show(
     root: Root,
-    bake_id: str,
+    bake: str,
     attempt: int,
     task_id: str,
     raw: bool,
@@ -210,16 +287,17 @@ async def show(
     async with AsyncExitStack() as stack:
         client = await stack.enter_async_context(neuro_sdk.get())
         storage: Storage = await stack.enter_async_context(
-            FSStorage(NeuroStorageFS(client))
+            APIStorage(client, NeuroStorageFS(client))
         )
         runner = await stack.enter_async_context(
-            BatchRunner(root.config_dir, root.console, client, storage)
+            BatchRunner(root.config_dir, root.console, client, storage, root)
         )
+        bake_id = await resolve_bake(bake, project=runner.project_id, storage=storage)
         await runner.logs(bake_id, task_id, attempt_no=attempt, raw=raw)
 
 
 @click.command()
-@argument("bake_id", type=BAKE)
+@argument("bake", type=BAKE)
 @click.option(
     "-a",
     "--attempt",
@@ -230,7 +308,7 @@ async def show(
 @wrap_async()
 async def cancel(
     root: Root,
-    bake_id: str,
+    bake: str,
     attempt: int,
 ) -> None:
     """Cancel a bake.
@@ -240,11 +318,12 @@ async def cancel(
     async with AsyncExitStack() as stack:
         client = await stack.enter_async_context(neuro_sdk.get())
         storage: Storage = await stack.enter_async_context(
-            FSStorage(NeuroStorageFS(client))
+            APIStorage(client, NeuroStorageFS(client))
         )
         runner = await stack.enter_async_context(
-            BatchRunner(root.config_dir, root.console, client, storage)
+            BatchRunner(root.config_dir, root.console, client, storage, root)
         )
+        bake_id = await resolve_bake(bake, project=runner.project_id, storage=storage)
         await runner.cancel(bake_id, attempt_no=attempt)
 
 
@@ -264,10 +343,10 @@ async def clear_cache(
     async with AsyncExitStack() as stack:
         client = await stack.enter_async_context(neuro_sdk.get())
         storage: Storage = await stack.enter_async_context(
-            FSStorage(NeuroStorageFS(client))
+            APIStorage(client, NeuroStorageFS(client))
         )
         runner = await stack.enter_async_context(
-            BatchRunner(root.config_dir, root.console, client, storage)
+            BatchRunner(root.config_dir, root.console, client, storage, root)
         )
         if batch == "ALL":
             await runner.clear_cache(None)
@@ -276,7 +355,7 @@ async def clear_cache(
 
 
 @click.command()
-@argument("bake_id", type=BAKE)
+@argument("bake", type=BAKE)
 @option(
     "-a",
     "--attempt",
@@ -294,7 +373,7 @@ async def clear_cache(
 @wrap_async()
 async def restart(
     root: Root,
-    bake_id: str,
+    bake: str,
     attempt: int,
     from_failed: bool,
     local_executor: bool,
@@ -306,14 +385,32 @@ async def restart(
     async with AsyncExitStack() as stack:
         client = await stack.enter_async_context(neuro_sdk.get())
         storage: Storage = await stack.enter_async_context(
-            FSStorage(NeuroStorageFS(client))
+            APIStorage(client, NeuroStorageFS(client))
         )
         runner = await stack.enter_async_context(
-            BatchRunner(root.config_dir, root.console, client, storage)
+            BatchRunner(root.config_dir, root.console, client, storage, root)
         )
+        bake_id = await resolve_bake(bake, project=runner.project_id, storage=storage)
         await runner.restart(
             bake_id,
             attempt_no=attempt,
             from_failed=from_failed,
             local_executor=local_executor,
         )
+
+
+def _parse_date(value: Optional[str]) -> Optional[datetime]:
+    if value:
+        try:
+            return isoparse(value)
+        except ValueError:
+            try:
+                delta = parse_timedelta(value)
+                return datetime.now(timezone.utc) - delta
+            except click.UsageError:
+                raise ValueError(
+                    "Date should be either in ISO-8601 format or "
+                    "relative delta of form 1d2h3m4s"
+                )
+    else:
+        return None
