@@ -5,6 +5,7 @@ import aiohttp
 import secrets
 import sys
 import tarfile
+from aiohttp.web_exceptions import HTTPNotFound
 from io import StringIO, TextIOWrapper
 from neuro_sdk import Client
 from pathlib import PureWindowsPath
@@ -198,7 +199,9 @@ class LocalCL(StreamCL, abc.ABC):
                 if not ret.is_file():
                     raise ValueError(f"Flow {ret} is not a file")
                 return ret
-        raise ValueError(f"Flow {name} does not exist")
+        raise ValueError(
+            f"Config file for flow '{name}' not found " f"in {self._config_dir} folder"
+        )
 
     @asynccontextmanager
     async def flow_stream(self, name: str) -> AsyncIterator[TextIO]:
@@ -255,6 +258,12 @@ class LocalCL(StreamCL, abc.ABC):
             async with self._github_session.get(
                 url=f"https://api.github.com/repos/{repo}/tarball/{version}"
             ) as response:
+                if response.status == HTTPNotFound.status_code:
+                    raise ValueError(
+                        "Cannot fetch action: "
+                        f"either repository '{repo}' or tag '{version}' does not exist"
+                    )
+                response.raise_for_status()
                 async for chunk in response.content:
                     file.write(chunk)
             file.seek(0)
@@ -380,7 +389,7 @@ class ConfigsMeta:
 
     @classmethod
     def from_json(cls, data: Mapping[str, Any]) -> "ConfigsMeta":
-        if data["project_config"]:
+        if data.get("project_config"):
             project_config: Optional[ConfigOnStorage] = ConfigOnStorage.from_json(
                 data["project_config"]
             )
