@@ -35,6 +35,7 @@ from typing import (
     ClassVar,
     Dict,
     Generic,
+    Iterable,
     Iterator,
     List,
     Mapping,
@@ -112,9 +113,19 @@ class EvalError(Exception):
         self.end = end
 
     def __str__(self) -> str:
-        line = self.start.line
-        col = self.start.col
-        return str(self.args[0]) + f"\n  in line {line}, column {col}"
+        # For humans, line and columns are enumerated from 1, so we should add 1 here.
+        line = self.start.line + 1
+        col = self.start.col + 1
+        filename = self.start.filename
+        return str(self.args[0]) + f'\n  in "{filename}", line {line}, column {col}'
+
+
+class MultiEvalError(Exception):
+    def __init__(self, errors: Sequence[EvalError]):
+        self.errors = errors
+
+    def __str__(self) -> str:
+        return "\n".join(str(error) for error in self.errors)
 
 
 def parse_literal(arg: str, err_msg: str) -> LiteralT:
@@ -162,6 +173,13 @@ async def akeys(ctx: CallCtx, arg: TypeT) -> TypeT:
     return list(arg)  # type: ignore  # List[...] is implicitly converted to SequenceT
 
 
+async def values(ctx: CallCtx, arg: TypeT) -> TypeT:
+    # Trampoline for mapping.values, async is required for the sake of uniformness.
+    if not isinstance(arg, Mapping):
+        raise TypeError(f"values() requires a mapping, got {arg!r}")
+    return list(arg.values())  # type: ignore  # implicitly converted to SequenceT
+
+
 async def fmt(ctx: CallCtx, spec: str, *args: TypeT) -> str:
     # We need a trampoline since expression syntax doesn't support classes and named
     # argumens
@@ -229,6 +247,12 @@ async def upload(ctx: CallCtx, volume_ctx: ContainerT) -> ContainerT:
 
     if not isinstance(volume_ctx, VolumeCtx):
         raise ValueError("upload() argument should be volume")
+    await run_subproc(
+        "neuro",
+        "mkdir",
+        "--parents",
+        str(volume_ctx.remote.parent),
+    )
     await run_subproc(
         "neuro",
         "cp",
@@ -255,6 +279,38 @@ async def aupper(ctx: CallCtx, arg: TypeT) -> str:
     return arg.upper()
 
 
+async def astr(ctx: CallCtx, arg: TypeT) -> str:
+    # Async version of str(), async is required for the sake of uniformness.
+    return str(arg)
+
+
+async def replace(ctx: CallCtx, arg: TypeT, old: TypeT, new: TypeT) -> str:
+    # We need a trampoline since expression syntax doesn't support classes
+    if not isinstance(arg, str):
+        raise TypeError(f"replace() first argument should be a str, got {arg!r}")
+    if not isinstance(old, str):
+        raise TypeError(f"replace() second argument should be a str, got {old!r}")
+    if not isinstance(new, str):
+        raise TypeError(f"replace() third argument should be a str, got {new!r}")
+    return arg.replace(old, new)
+
+
+async def join(ctx: CallCtx, sep: TypeT, array: TypeT) -> str:
+    # We need a trampoline since expression syntax doesn't support classes
+    if not isinstance(sep, str):
+        raise TypeError(f"join() first argument should be a str, got {sep!r}")
+    if not isinstance(array, SequenceT):
+        raise TypeError(
+            f"replace() second argument should be a sequence, got {array!r}"
+        )
+    str_array = []
+    for idx, item in enumerate(array):
+        if not isinstance(item, str):
+            raise TypeError(f"join() array item {idx} should be a str, got {item!r}")
+        str_array.append(item)
+    return sep.join(str_array)
+
+
 async def parse_volume(ctx: CallCtx, arg: TypeT) -> ContainerT:
     from .context import VolumeCtx
 
@@ -290,7 +346,7 @@ def _get_needs_statuses(root: RootABC) -> Dict[str, TaskStatus]:
 
 
 async def always(ctx: CallCtx, *args: str) -> AlwaysT:
-    _check_has_needs(ctx, func_name="success")
+    _check_has_needs(ctx, func_name="always")
     return AlwaysT()
 
 
@@ -376,7 +432,11 @@ FUNCTIONS = _build_signatures(
     len=alen,
     nothing=nothing,
     fmt=fmt,
+    str=astr,
+    replace=replace,
+    join=join,
     keys=akeys,
+    values=values,
     to_json=to_json,
     from_json=from_json,
     success=success,
@@ -402,6 +462,9 @@ class Item(Entity):
     async def eval(self, root: RootABC) -> TypeT:
         pass
 
+    def child_items(self) -> Iterable["Item"]:
+        return []
+
 
 @dataclasses.dataclass(frozen=True)
 class Literal(Item):
@@ -485,6 +548,11 @@ class Lookup(Item):
             ret = await op.eval(root, ret, start)
         return ret
 
+    def child_items(self) -> Iterable["Item"]:
+        for getter in self.trailer:
+            if isinstance(getter, ItemGetter):
+                yield getter.key
+
 
 def make_lookup(arg: Tuple[Token, List[Getter]]) -> Lookup:
     name, trailer = arg
@@ -522,6 +590,12 @@ class Call(Item):
             ret = await op.eval(root, ret, start)
         return ret
 
+    def child_items(self) -> Iterable["Item"]:
+        yield from self.args
+        for getter in self.trailer:
+            if isinstance(getter, ItemGetter):
+                yield getter.key
+
 
 def make_call(arg: Tuple[Token, List[Item], Sequence[Getter]]) -> Call:
     funcname, args, trailer = arg
@@ -563,6 +637,9 @@ class BinOp(Item):
         right_val = await self.right.eval(root)
         return self.op(left_val, right_val)  # type: ignore
 
+    def child_items(self) -> Iterable["Item"]:
+        return [self.left, self.right]
+
 
 def or_(arg1: Any, arg2: Any) -> Any:
     # Emulate dicts concatination from Python 3.9
@@ -635,6 +712,9 @@ class UnaryOp(Item):
         operand_val = await self.operand.eval(root)
         return self.op(operand_val)  # type: ignore
 
+    def child_items(self) -> Iterable["Item"]:
+        return [self.operand]
+
 
 def _unary_plus(arg: Any) -> Any:
     return +arg
@@ -666,12 +746,19 @@ class ListMaker(Item):
     async def eval(self, root: RootABC) -> SequenceT:
         return [await item.eval(root) for item in self.items]  # type: ignore
 
+    def child_items(self) -> Iterable["Item"]:
+        return self.items
+
 
 def make_list(args: Tuple[Item, List[Item]]) -> ListMaker:
     lst = [args[0]] + args[1]
     return ListMaker(lst[0].start, lst[-1].end, lst)
 
 
+def make_empty_list(args: Tuple[Item, Item]) -> ListMaker:
+    return ListMaker(args[0].start, args[1].end, [])
+
+
 @dataclasses.dataclass(frozen=True)
 class DictMaker(Item):
     items: Sequence[Tuple[Item, Item]]
@@ -684,6 +771,11 @@ class DictMaker(Item):
             ret[k] = v
         return ret  # type: ignore
 
+    def child_items(self) -> Iterable["Item"]:
+        for entry in self.items:
+            yield entry[0]
+            yield entry[1]
+
 
 def make_dict(args: Tuple[Item, Item, List[Tuple[Item, Item]]]) -> DictMaker:
     lst = [(args[0], args[1])]
@@ -692,6 +784,10 @@ def make_dict(args: Tuple[Item, Item, List[Tuple[Item, Item]]]) -> DictMaker:
     return DictMaker(lst[0][0].start, lst[-1][1].end, lst)
 
 
+def make_empty_dict(args: Tuple[Item, Item]) -> DictMaker:
+    return DictMaker(args[0].start, args[0].end, [])
+
+
 def a(value: str) -> Parser:
     """Eq(a) -> Parser(a, a)
 
@@ -825,7 +921,10 @@ FACTOR.define((OP_PLUS | OP_MINUS) + FACTOR >> make_unary_op_expr | ATOM_EXPR)
 
 EXPR.define(DISJUNCTION)  # Just synonym
 
-LIST_MAKER.define((LSQB + EXPR + many(COMMA + EXPR) + maybe(COMMA) + RSQB) >> make_list)
+LIST_MAKER.define(
+    (LSQB + EXPR + many(COMMA + EXPR) + maybe(COMMA) + RSQB) >> make_list
+    | (a("[") + a("]")) >> make_empty_list
+)
 
 DICT_MAKER.define(
     (
@@ -838,6 +937,7 @@ DICT_MAKER.define(
         + RBRACE
     )
     >> make_dict
+    | (a("{") + a("}")) >> make_empty_dict
 )
 
 TMPL: Final = (OPEN_TMPL + EXPR + CLOSE_TMPL) | (OPEN_TMPL2 + EXPR + CLOSE_TMPL2)
@@ -860,6 +960,8 @@ class Expr(BaseExpr[_T]):
     allow_none: ClassVar[bool] = True
     allow_expr: ClassVar[bool] = True
     type: ClassVar[Type[_T]]
+    start: Pos
+    end: Pos
     _ret: Union[None, _T]
     _pattern: Union[None, str, _T]
     _parsed: Optional[Sequence[Item]]
@@ -875,6 +977,8 @@ class Expr(BaseExpr[_T]):
             raise EvalError(str(exc), start, end)
 
     def __init__(self, start: Pos, end: Pos, pattern: Union[None, str, _T]) -> None:
+        self.start = start
+        self.end = end
         self._pattern = pattern
         # precalculated value for constant string, allows raising errors earlier
         self._ret = None
@@ -939,7 +1043,7 @@ class Expr(BaseExpr[_T]):
                 except EvalError:
                     raise
                 except Exception as exc:
-                    raise EvalError(str(exc), part.start, part.end)
+                    raise EvalError(str(exc), part.start, part.end) from exc
                 # TODO: add str() function, raise an explicit error if
                 # an expresion evaluates non-str type
                 # assert isinstance(val, str), repr(val)
