@@ -5,7 +5,7 @@ from neuro_flow import ast
 from neuro_flow.ast import BatchActionOutputs
 from neuro_flow.expr import (
     EnableExpr,
-    IdExpr,
+    MappingItemsExpr,
     OptBashExpr,
     OptBoolExpr,
     OptIdExpr,
@@ -13,6 +13,7 @@ from neuro_flow.expr import (
     OptRemotePathExpr,
     OptStrExpr,
     OptTimeDeltaExpr,
+    SequenceItemsExpr,
     SimpleOptBoolExpr,
     SimpleOptIdExpr,
     SimpleOptStrExpr,
@@ -120,11 +121,11 @@ def test_params_forbidden_in_live_action(assets: LocalPath) -> None:
 
 
 def test_parse_batch_action(assets: LocalPath) -> None:
-    config_file = assets / "batch-action.yml"
+    config_file = assets / "batch-action-with-image.yml"
     action = parse_action(config_file)
     assert action == ast.BatchAction(
         Pos(0, 0, config_file),
-        Pos(28, 0, config_file),
+        Pos(43, 0, config_file),
         kind=ast.ActionKind.BATCH,
         name=SimpleOptStrExpr(
             Pos(0, 0, config_file),
@@ -166,10 +167,6 @@ def test_parse_batch_action(assets: LocalPath) -> None:
         outputs=BatchActionOutputs(
             Pos(ANY, ANY, config_file),
             Pos(ANY, ANY, config_file),
-            needs=[
-                IdExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), "task_1"),
-                IdExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), "task_2"),
-            ],
             values={
                 "res1": ast.Output(
                     Pos(ANY, ANY, config_file),
@@ -209,13 +206,65 @@ def test_parse_batch_action(assets: LocalPath) -> None:
                 Pos(0, 0, config_file), Pos(0, 0, config_file), "30m"
             ),
         ),
+        images={
+            "image_a": ast.Image(
+                _start=Pos(23, 4, config_file),
+                _end=Pos(35, 0, config_file),
+                ref=StrExpr(
+                    Pos(0, 0, config_file), Pos(0, 0, config_file), "image:banana"
+                ),
+                context=OptStrExpr(
+                    Pos(0, 0, config_file), Pos(0, 0, config_file), "dir"
+                ),
+                dockerfile=OptStrExpr(
+                    Pos(0, 0, config_file), Pos(0, 0, config_file), "dir/Dockerfile"
+                ),
+                build_args=SequenceItemsExpr(
+                    [
+                        StrExpr(
+                            Pos(0, 0, config_file), Pos(0, 0, config_file), "--arg1"
+                        ),
+                        StrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), "val1"),
+                        StrExpr(
+                            Pos(0, 0, config_file),
+                            Pos(0, 0, config_file),
+                            "--arg2=val2",
+                        ),
+                    ]
+                ),
+                env=MappingItemsExpr(
+                    {
+                        "SECRET_ENV": StrExpr(
+                            Pos(0, 0, config_file), Pos(0, 0, config_file), "secret:key"
+                        ),
+                    }
+                ),
+                volumes=SequenceItemsExpr(
+                    [
+                        OptStrExpr(
+                            Pos(0, 0, config_file),
+                            Pos(0, 0, config_file),
+                            "secret:key:/var/secret/key.txt",
+                        ),
+                    ]
+                ),
+                build_preset=OptStrExpr(
+                    Pos(0, 0, config_file), Pos(0, 0, config_file), "gpu-small"
+                ),
+                force_rebuild=OptBoolExpr(
+                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
+                ),
+            )
+        },
         tasks=[
             ast.Task(
-                Pos(22, 2, config_file),
-                Pos(25, 0, config_file),
+                Pos(36, 2, config_file),
+                Pos(40, 0, config_file),
                 title=OptStrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), None),
                 name=OptStrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), None),
-                image=StrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), "ubuntu"),
+                image=StrExpr(
+                    Pos(0, 0, config_file), Pos(0, 0, config_file), "image:banana"
+                ),
                 preset=OptStrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), None),
                 schedule_timeout=OptTimeDeltaExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), None
@@ -247,7 +296,7 @@ def test_parse_batch_action(assets: LocalPath) -> None:
                     Pos(0, 0, config_file), Pos(0, 0, config_file), None
                 ),
                 id=OptIdExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), "task_1"),
-                needs=None,
+                needs={},
                 strategy=None,
                 cache=None,
                 enable=EnableExpr(
@@ -257,8 +306,8 @@ def test_parse_batch_action(assets: LocalPath) -> None:
                 ),
             ),
             ast.Task(
-                Pos(25, 2, config_file),
-                Pos(28, 0, config_file),
+                Pos(40, 2, config_file),
+                Pos(43, 0, config_file),
                 title=OptStrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), None),
                 name=OptStrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), None),
                 image=StrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), "ubuntu"),
@@ -484,6 +533,9 @@ def test_parse_batch_call(assets: LocalPath) -> None:
             Pos(0, 0, config_file),
             None,
         ),
+        life_span=OptTimeDeltaExpr(
+            Pos(0, 0, config_file), Pos(0, 0, config_file), None
+        ),
         images=None,
         volumes=None,
         defaults=None,
