@@ -205,7 +205,7 @@ def test_parse_full(assets: pathlib.Path) -> None:
     flow = parse_live(workspace, config_file)
     assert flow == ast.LiveFlow(
         Pos(0, 0, config_file),
-        Pos(61, 0, config_file),
+        Pos(63, 0, config_file),
         id=SimpleOptIdExpr(
             Pos(0, 0, config_file),
             Pos(0, 0, config_file),
@@ -224,10 +224,10 @@ def test_parse_full(assets: pathlib.Path) -> None:
                 ref=StrExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), "image:banana"
                 ),
-                context=OptLocalPathExpr(
+                context=OptStrExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), "dir"
                 ),
-                dockerfile=OptLocalPathExpr(
+                dockerfile=OptStrExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), "dir/Dockerfile"
                 ),
                 build_args=SequenceItemsExpr(
@@ -262,6 +262,9 @@ def test_parse_full(assets: pathlib.Path) -> None:
                 build_preset=OptStrExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), "gpu-small"
                 ),
+                force_rebuild=OptBoolExpr(
+                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
+                ),
             )
         },
         volumes={
@@ -300,7 +303,7 @@ def test_parse_full(assets: pathlib.Path) -> None:
         },
         defaults=ast.FlowDefaults(
             Pos(26, 2, config_file),
-            Pos(34, 0, config_file),
+            Pos(36, 0, config_file),
             tags=SequenceItemsExpr(
                 [
                     StrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), "tag-a"),
@@ -317,6 +320,15 @@ def test_parse_full(assets: pathlib.Path) -> None:
                     ),
                 }
             ),
+            volumes=SequenceItemsExpr(
+                [
+                    OptStrExpr(
+                        Pos(0, 0, config_file),
+                        Pos(0, 0, config_file),
+                        "storage:common:/mnt/common:rw",
+                    ),
+                ]
+            ),
             workdir=OptRemotePathExpr(
                 Pos(0, 0, config_file), Pos(0, 0, config_file), "/global/dir"
             ),
@@ -332,8 +344,8 @@ def test_parse_full(assets: pathlib.Path) -> None:
         ),
         jobs={
             "test_a": ast.Job(
-                Pos(36, 4, config_file),
-                Pos(61, 0, config_file),
+                Pos(38, 4, config_file),
+                Pos(63, 0, config_file),
                 name=OptStrExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), "job-name"
                 ),
@@ -444,7 +456,7 @@ def test_parse_full_exprs(assets: pathlib.Path) -> None:
     flow = parse_live(workspace, config_file)
     assert flow == ast.LiveFlow(
         Pos(0, 0, config_file),
-        Pos(47, 0, config_file),
+        Pos(48, 0, config_file),
         id=SimpleOptIdExpr(
             Pos(0, 0, config_file),
             Pos(0, 0, config_file),
@@ -463,10 +475,10 @@ def test_parse_full_exprs(assets: pathlib.Path) -> None:
                 ref=StrExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), "image:banana"
                 ),
-                context=OptLocalPathExpr(
+                context=OptStrExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), "dir"
                 ),
-                dockerfile=OptLocalPathExpr(
+                dockerfile=OptStrExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), "dir/Dockerfile"
                 ),
                 build_args=SequenceExpr(
@@ -490,6 +502,9 @@ def test_parse_full_exprs(assets: pathlib.Path) -> None:
                 build_preset=OptStrExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), "gpu-small"
                 ),
+                force_rebuild=OptBoolExpr(
+                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
+                ),
             )
         },
         volumes={
@@ -528,7 +543,7 @@ def test_parse_full_exprs(assets: pathlib.Path) -> None:
         },
         defaults=ast.FlowDefaults(
             Pos(21, 2, config_file),
-            Pos(27, 0, config_file),
+            Pos(28, 0, config_file),
             tags=SequenceExpr(
                 Pos(0, 0, config_file),
                 Pos(0, 0, config_file),
@@ -541,6 +556,12 @@ def test_parse_full_exprs(assets: pathlib.Path) -> None:
                 "${{ {'global_a': 'val-a', 'global_b': 'val-b'} }}",
                 type2str,
             ),
+            volumes=SequenceExpr(
+                Pos(0, 0, config_file),
+                Pos(0, 0, config_file),
+                "${{ ['storage:common:/mnt/common:rw'] }}",
+                type2str,
+            ),
             workdir=OptRemotePathExpr(
                 Pos(0, 0, config_file), Pos(0, 0, config_file), "/global/dir"
             ),
@@ -556,8 +577,8 @@ def test_parse_full_exprs(assets: pathlib.Path) -> None:
         ),
         jobs={
             "test_a": ast.Job(
-                Pos(29, 4, config_file),
-                Pos(47, 0, config_file),
+                Pos(30, 4, config_file),
+                Pos(48, 0, config_file),
                 name=OptStrExpr(
                     Pos(0, 0, config_file), Pos(0, 0, config_file), "job-name"
                 ),
@@ -825,9 +846,9 @@ def test_bad_expr_type_before_eval(assets: pathlib.Path) -> None:
     with pytest.raises(EvalError) as ctx:
         parse_live(workspace, config_file)
     assert str(ctx.value) == dedent(
-        """\
+        f"""\
         invalid literal for int() with base 10: 'abc def'
-          in line 5, column 15"""
+          in "{config_file}", line 6, column 16"""
     )
 
 
